"""
Configuring Runestone Servers
=============================

Many things about Runestone servers are configurable. This is the place to change
the configuration for most things.  **Private** things should be configured in the
environment so they are not accidentally committed to Github.
Defaults provided here may be overridden by environment variables `Per <https://fastapi.tiangolo.com/advanced/settings/>`_.
"""

#
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8`_.
#
# Standard library
# ----------------
from enum import Enum
from functools import lru_cache
import os
from pathlib import Path, PosixPath

# Local application imports
# -------------------------
from rsptx.logging import rslogger
from pydantic_settings import BaseSettings, SettingsConfigDict

# Settings
# ========
# Define the possible bookserver configurations. The values assigned must  be strings, since Pydantic will match these with environment variables.


class BookServerConfig(Enum):
    development = "development"
    test = "test"
    production = "production"


# A enum for the type of database in use.
class DatabaseType(Enum):
    SQLite = 0
    PostgreSQL = 1


class Settings(BaseSettings):
    """
    Pydantic provides a wonderful utility to handle settings.  The beauty of it
    is that you can specify variables with or without default values, and Pydantic
    will check your environment variables in a case-insensitive way. So that
    if you have ``PROD_DBURL``` set in the environment it will be set as the value
    for ``prod_dburl``` in settings.
    This is a really nice way to keep from
    committing any data you want to keep private.
    """

    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
        extra="allow",
    )

    google_ga: str = ""

    # The path to the Runestone application inside web2py.

    runestone_path: PosixPath = Path(
        os.environ.get("RUNESTONE_PATH", Path.home())
    ).resolve()

    # _`book_path`: specify the directory to serve books from. For now, default to serving from the same place as the Runestone server, since the server uses some of these files.
    book_path: Path = runestone_path / "books"

    # The path to store error logs.
    error_path: Path = Path(os.environ.get("BOOK_PATH", "/usr/books")) / "tickets"

    # Define the mode of operation for the webserver, taken from ``BookServerConfig```. This looks a bit odd, since the string value will be parsed by Pydantic into a Config.
    #
    # .. admonition:: warning
    #
    #    When using an Enum for a configuration setting you cannot compare against
    #    a string.  The value will actually be ``BookServerConfig.development``` or whatever.
    #    Our style will be to compare against the Enum not the ``.name`` attribute.
    #

    server_config: str = "development"

    # server_config is the master make sure everything else matches it
    # can be overriden from the Settings constructor. which we do using
    # the server_config environment variable
    book_server_config: BookServerConfig = BookServerConfig.development

    # Database setup: this must be an standard (synchronous) connection; for example:
    #
    # - ``sqlite:///./runestone.db``
    # - ``postgresql://postgres:bully@localhost/runestone``
    dburl: str = f"sqlite:///{runestone_path}/runestone.db"
    dev_dburl: str = f"sqlite:///{runestone_path}/runestone_dev.db"
    test_dburl: str = f"sqlite:///{runestone_path}/runestone_test.db"

    # Given a sync database URI, return its async equivalent.
    @staticmethod
    def _sync_to_async_uri(sync_uri: str) -> str:
        return sync_uri.replace("sqlite://", "sqlite+aiosqlite://", 1).replace(
            "postgresql://", "postgresql+asyncpg://", 1
        )

    # Determine the synchronous database URL based on the ``book_server_config`` and the dburls above.
    @property
    def _sync_database_url(self) -> str:
        return {
            "development": self.dev_dburl,
            "test": self.test_dburl,
            "production": self.dburl,
        }[self.server_config]

    # Return the async equivalent of the URI from ``sync_database_url``.
    @property
    def database_url(self) -> str:
        return self._sync_to_async_uri(self._sync_database_url)

    # Determine the database type from the URL.
    @property
    def database_type(self) -> DatabaseType:
        dburl = self.database_url
        if dburl.startswith("sqlite"):
            return DatabaseType.SQLite
        elif dburl.startswith("postgresql"):
            return DatabaseType.PostgreSQL
        else:
            raise RuntimeError(f"Unknown database type; URL is {dburl}.")

    # Setting db_echo to True makes for a LOT of sqlalchemy output - it gives you the SQL for every query!
    db_echo: bool = False

    # The docker-compose.yml file will set the REDIS_URI environment variable
    redis_uri: str = "redis://localhost:6379/0"

    # Select normal mode or a high-stakes assessment mode (for administering a examination). In this mode, answers to supported question types are not shown.
    is_exam: bool = False

    # This module provides "/auth/login" but it is super basic
    # For now, logins are meant to be handled by the parallel runestone server.
    # For production usage set the LOGIN_URL to /runestone/default/user
    login_url: str = "/auth/login"

    # Configure ads. TODO: Link to the place in the Runestone Components where this is used.
    adsenseid: str = ""
    num_banners: int = 0
    serve_ad: bool = False
    academy_mode: bool = True
    lti_only_mode: bool = False

    # This is the secret key used for generating the JWT token.
    jwt_secret: bytes = b"supersecret"

    # This is the private key web2py uses for hashing passwords.
    @property
    def web2py_private_key(self) -> str:
        """Get the web2py private key.
        Prefer to get it from the environment.  This allows us to avoid introducing a secret
        into the codebase, and it also avoids the need to mount a volume to the container to
        store the key in a file.

        :return: The web2py private key.
        :rtype: str
        """

        # Put the cache here; above the def, it produces ``TypeError: unhashable type: 'Settings'``.
        @lru_cache
        def read_key():
            key_file = Path(self.runestone_path, "private", "auth.key")
            if key_file.exists():
                with open(key_file, encoding="utf-8") as f:
                    return f.read().strip()
            else:
                rslogger.warning(
                    "No Key file OR WEB2PY_PRIVATE_KEY -  will default to settings.jwt_secret"
                )
                if type(self.jwt_secret) is bytes:
                    return self.jwt_secret.decode("utf-8")
                return self.jwt_secret

        key = os.environ.get("WEB2PY_PRIVATE_KEY", None)
        if key:
            return key
        else:
            return read_key()

    # if you want to reinitialize your database set this to ``Yes``.
    # All data in the database will be lost! This will only work for
    # development and test ``book_server_config`` settings.
    drop_tables: str = "No"

    # needed for using Spaces / AWS S3 for file uploads
    # see
    spaces_key: str = "key"
    spaces_secret: str = "secret"
    region: str = "nyc3"  # this is the DO data center or AWS region
    bucket: str = "runestonefiles"

    log_level: str = "DEBUG"

    jobe_key: str = ""
    jobe_server: str = "http://jobe"


settings = Settings(book_server_config=os.environ.get("SERVER_CONFIG", "development"))



# Common routines for converting to PreTeXt


def write_substitute(self, node, html=None):
    self.output.append(
        "<exercise runestone='{divid}' />".format(**node["runestone_options"]))
    with open("rs-substitutes.xml", "a") as subfile:
        subfile.write("<substitute xml:id='{divid}'>".format(
            **node["runestone_options"]))
        if html:
            subfile.write(f"""{html}""")
        else:
            subfile.write(
                f"""<dbfetch>{node["runestone_options"]["divid"]}</dbfetch>""")
        subfile.write("</substitute>")


def substitute_visitor(self, node):
    write_substitute(self, node)


def substitute_departure(self, node):
    pass



# *******************************************
# |docname| - reusable functions for rsmanage
# *******************************************
# These functions are used by the rsmanage command in RunestoneServer as well as
# by the AuthorServer in its Celery worker tasks.  There may be other places that
# find these utils handy as well.
#
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8`_.
#
# Standard library
# ----------------
import datetime
import os
import re
import subprocess
from pathlib import Path
import logging
from io import StringIO
from shutil import copytree

# Third Party
# -----------
import click
import lxml.etree as ET
from lxml import ElementInclude
import pretext
from pretext.utils import is_earlier_version
import pretext.project

# import xml.etree.ElementTree as ET

from sqlalchemy import create_engine, Table, MetaData, and_, update
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.sql import text

# todo: use our logger
from rsptx.logging import rslogger
from runestone.server import get_dburl
from rsptx.db.models import Library, LibraryValidator
from rsptx.response_helpers.core import canonical_utcnow
import pdb

rslogger.setLevel("WARNING")

# Local packages
# --------------


QT_MAP = {
    "multiplechoice": "mchoice",
    "parsons": "parsonsprob",
}


# Build a Runestone Book
# ----------------------
def _build_runestone_book(config, course, click=click):
    """
    Parameters:
    course: the name of the course to build.
    click: default is the click module otherwise an object that has an echo method
    """
    try:
        if os.path.exists("pavement.py"):
            # Since this may be used in a long running process (see AuthorServer/worker)
            # using import is a bad idea, exec can be dangerous as well
            paver_vars = {}
            exec(open("pavement.py").read(), paver_vars)
        else:
            click.echo(
                f"I can't find a pavement.py file in {os.getcwd()} you need that to build"
            )
            return False
    except ImportError as e:
        click.echo("You do not appear to have a good pavement.py file.")
        print(e)
        return False

    # If the click object has a worker attribute then we are running in a worker
    # process and **know** we are making a build for a runestone server. In that
    # case we need to make sure that the dynamic_pages flag is set to True in
    # pavement.py
    if hasattr(click, "worker") and paver_vars["dynamic_pages"] is not True:
        click.echo("dynamic_pages must be set to True in pavement.py")
        return False
    if paver_vars["project_name"] != course:
        click.echo(
            f"Error: {course} and {paver_vars['project_name']} do not match.  Your course name needs to match the project_name in pavement.py"
        )
        return False
    if paver_vars["options"].build.template_args["basecourse"] != course:
        click.echo(
            f"Error: {course} and {paver_vars['options'].build.template_args['basecourse']} do not match.  Your course name needs to match the basecourse in pavement.py"
        )
        return False

    click.echo("Running runestone build --all")
    res = subprocess.run("runestone build --all", shell=True, capture_output=True)
    with open("cli.log", "wb") as olfile:
        olfile.write(res.stdout)
        olfile.write(b"\n====\n")
        olfile.write(res.stderr)

    if res.returncode != 0:
        click.echo(
            f"building the book failed {res}, check the log for errors and try again"
        )
        return False
    click.echo("Build succeedeed... Now deploying to published")
    if paver_vars["dest"] != "./published":
        click.echo(
            "Incorrect deployment directory.  dest should be ./published in pavement.py"
        )
        return False

    resd = subprocess.run("runestone deploy", shell=True, capture_output=True)
    with open("cli.log", "ab") as olfile:
        olfile.write(res.stdout)
        olfile.write(b"\n====\n")
        olfile.write(res.stderr)
    if resd.returncode == 0:
        click.echo("Success! Book deployed")
    else:
        click.echo("Deploy failed, check the log to see what went wrong.")
        return False

    update_library(config, "", course, click, build_system="Runestone")
    return True


# Build a PreTeXt Book
# --------------------
def _build_ptx_book(config, gen, manifest, course, click=click, target="runestone"):
    """
    Parameters:
    config : This originated as a config object from click -- a mock config will be provided by the AuthorServer
    gen: A flag to indicate whether or not we should build static assets
    manifest: the name of the manifest file
    course: the name of the course to build.
    click: default is the click module otherwise an object that has an echo method
    """

    if not os.path.exists("project.ptx"):
        click.echo("PreTeXt books need a project.ptx file")
        return False
    else:
        click.echo("Checking files")
        if not target:
            target = "runestone"
        # sets output_dir to `published/<course>`
        # and {"host-platform": "runestone"} in stringparams
        rs = check_project_ptx(course=course, target=target)
        if not rs:
            return False

        logger = logging.getLogger("ptxlogger")
        string_io_handler = StringIOHandler()
        logger.addHandler(string_io_handler)
        click.echo("Building the book")

        rs.build()  # build the book, generating assets as needed
        log_path = Path(os.environ.get("BOOK_PATH")) / rs.output_dir / "cli.log"
        if not log_path.parent.exists():
            log_path.parent.mkdir(parents=True, exist_ok=True)
        click.echo(f"Writing log to {log_path}")
        with open(log_path, "a") as olfile:
            olfile.write(string_io_handler.getvalue())

        book_path = (
            Path(os.environ.get("BOOK_PATH"))
            / rs.output_dir
            / "published"
            / rs.output_dir
        )

        click.echo(f"Book will be deployed to {book_path}")
        if rs.output_dir_abspath() != book_path:
            res = copytree(rs.output_dir_abspath(), book_path, dirs_exist_ok=True)
            if not res:
                click.echo("Error copying files to published")
                return False
        else:
            click.echo("No need to copy files to published")
        click.echo("Book deployed successfully")
        mpath = rs.output_dir_abspath() / manifest
        process_manifest(course, mpath)
        # Fetch and copy the runestone components release as advertised by the manifest
        # - Use wget to get all the js files and put them in _static
        # Beginning with 2.6.1 PreTeXt populates the _static folder with the latest
        if is_earlier_version(pretext.VERSION, "2.6.1"):
            click.echo("populating with the latest runestone files")
            populate_static(config, mpath, course)
        # update the library page
        click.echo("updating library metadata...")
        main_page = find_real_url(course)
        update_library(config, mpath, course, main_page=main_page, build_system="PTX")
        return True


# Support Functions
# -----------------


def process_manifest(cname, mpath, click=click):
    """
    cname - the name of the course
    mpath - path to the runestone-manifest.xml file

    Setup this book in the database and populate the questions table as well as
    The chapter and subchapter tables.
    """
    click.echo("processing manifest...")
    if os.path.exists(mpath):
        manifest_data_to_db(cname, mpath)
    else:
        raise IOError(
            f"You must provide a valid path to a manifest file: {mpath} does not exist."
        )
    return True


def check_project_ptx(click=click, course=None, target="runestone"):
    """
    Verify that the PreTeXt project is set up for a Runestone build

    Returns: Runestone target from PreTeXt project

    1. Ensure there is a runestone target in project.ptx
    2. Set project output to published directory
    3. Ensure the top level source file exists
    4. Ensure the publisher file exists
    5. Ensure the document-id exists
    6. Set target output to document-id

    """
    proj = pretext.project.Project.parse("project.ptx")
    if not target:
        target = "runestone"
    target_name = target
    if proj.has_target(target_name) is False:
        if proj.has_target("web"):
            target_name = "web"
        elif proj.has_target("html"):
            target_name = "html"
        else:
            click.echo(f"No {target} suitable targets in project.ptx")
            return False
        click.echo(
            f"No {target} target in project.ptx, will adopt {target_name} target"
        )

    book_path = os.environ.get("BOOK_PATH", None)
    if book_path is None:
        click.echo("BOOK_PATH must be set in the environment")
        return False

    tgt = proj.get_target(target_name)

    rslogger.info(f"target name: {target_name}")
    rslogger.info(f"target source: {tgt.source_abspath()}")
    rslogger.info(f"target publication: {tgt.publication_abspath()}")
    if not tgt.source_abspath().exists():
        click.echo(f"Source file specified in {target_name} target does not exist")
        return False
    if not tgt.publication_abspath().exists():
        click.echo(f"Publication file specified in {target_name} target does not exist")
        return False

    docid_list = tgt.source_element().xpath("/pretext/docinfo/document-id/text()")
    if len(docid_list) < 1:
        click.echo(
            f"Source file specified in runestone {target_name} does not have a document-id"
        )
        docid = course
    else:
        docid = docid_list[0]

    if course is not None and docid != course:
        click.echo(f"Error course: {course} does not match document-id: {docid}")
        return False

    tgt.output_dir = Path(docid)

    tgt.stringparams.update({"host-platform": "runestone"})

    return tgt


def extract_docinfo(tree, string, attr=None, click=click):
    """
    Parameters:
    tree: The parsed document tree from ET
    string: The name of the element we are looking for
    Helper to get the contents of several tags from the docinfo element of a PreTeXt book
    """
    el = tree.find(f"./{string}")
    if attr is not None and el is not None:
        print(f"{el.attrib[attr]=}")
        return el.attrib[attr].strip()

    if el is not None:
        # using method="text" will strip the outer tag as well as any html tags in the value
        return ET.tostring(el, encoding="unicode", method="text").strip()
    return ""


def update_library(
    config, mpath, course, click=click, build_system="", main_page="index.html"
):
    """
    Parameters:
    config : This originated as a config object from click -- a mock config will be provided by the AuthorServer
    mpath: Path to the runestone-manifest file which containes the library metadata
    course: the name of the course we are buildingn

    Update the library table using meta data from the book

    Returns: Nothing
    """
    # This is a bit of a hack for now... todo: continue to refactor these to use crud functions
    eng = create_engine(config.dburl.replace("+asyncpg", ""))
    if build_system == "PTX":
        tree = ET.parse(mpath)
        docinfo = tree.find("./library-metadata")
        title = extract_docinfo(docinfo, "title")
        subtitle = extract_docinfo(docinfo, "subtitle")
        description = extract_docinfo(docinfo, "blurb")
        shelf = extract_docinfo(docinfo, "shelf")
    else:
        try:
            config_vars = {}
            exec(open("conf.py").read(), config_vars)
        except Exception as e:
            print(f"Error adding book {course} to library list: {e}")
            return
        subtitle = ""
        if "navbar_title" in config_vars:
            title = config_vars["navbar_title"]
        elif "html_title" in config_vars:
            title = config_vars["html_title"]
        elif "html_short_title" in config_vars:
            title = config_vars["html_short_title"]
        else:
            title = "Runestone Book"
        # update course description if found in the book's conf.py
        if "course_description" in config_vars:
            description = config_vars["course_description"]
        else:
            description = ""
        # update course key_words if found in book's conf.py
        # if "key_words" in config_vars:
        #     key_words = config_vars["key_words"]

        if "shelf_section" in config_vars:
            shelf = config_vars["shelf_section"]
        else:
            shelf = "Computer Science"

    click.echo(f"{title} : {subtitle}")

    Session = sessionmaker()
    eng.connect()
    Session.configure(bind=eng)

    try:
        res = eng.execute(f"select * from library where basecourse = '{course}'")
    except Exception:
        click.echo("Missing library table?  You may need to run an alembic migration.")
        return False
    # using the Model rather than raw sql ensures that everything is properly escaped
    build_time = canonical_utcnow()
    click.echo(f"BUILD time is {build_time}")
    if res.rowcount == 0:
        new_lib = LibraryValidator(
            title=title,
            subtitle=subtitle,
            description=description,
            shelf_section=shelf,
            basecourse=course,
            build_system=build_system,
            main_page=main_page,
            last_build=build_time,
            for_classes="F",
            is_visible="T",
        )
        new_book = Library(**new_lib.dict())
        with Session.begin() as s:
            s.add(new_book)
    else:
        # If any values are missing or null do not override them here.
        #
        res = res.first()
        if not title:
            title = res.title or ""
        if not subtitle:
            subtitle = res.subtitle or ""
        if not description:
            description = res.description or ""
        if not shelf:
            shelf = res.shelf_section or "Misc"
        click.echo("Updating library")
        stmt = (
            update(Library)
            .where(Library.basecourse == course)
            .values(
                title=title,
                subtitle=subtitle,
                description=description,
                shelf_section=shelf,
                build_system=build_system,
                main_page=main_page,
                last_build=build_time,
            )
        )
        with Session.begin() as session:
            session.execute(stmt)
    return True


def find_real_url(book):
    idx = Path("published", book, "index.html")
    if idx.exists():
        with open(idx, "r") as idxf:
            for line in idxf:
                if g := re.search(r"refresh.*URL='(.*?)'", line):
                    return g.group(1)
    return "index.html"


def populate_static(config, mpath: Path, course: str, click=click):
    """
    Copy the apropriate Javascript to the _static folder for PreTeXt books.  This may
    involve downloading it from the Runestone CDN.  PreTeXt does not include the current set
    of javascript files like the Runestone components release does, instead we supply it
    on runestone.academy/cdn/runestone so it can be used for generic html builds as well as
    builds on runestone.academy.
    """
    # <runestone-services version="6.2.1"/>
    sdir = mpath.parent / "_static"
    current_version = ""
    if (sdir / "webpack_static_imports.xml").exists():
        tree = ET.parse(sdir / "webpack_static_imports.xml")
        current_version = tree.find("./version").text
    else:
        sdir.mkdir(mode=0o775, exist_ok=True)  # NB mode must be in Octal!
    if mpath.exists():
        tree = ET.parse(mpath)
        el = tree.find("./runestone-services[@version]")
        version = el.attrib["version"].strip()
    else:
        click.echo("Error: missing runestone-manifest.xml file")
        return False
    # Do not download if the versions already match.
    if version != current_version:
        click.echo(f"Fetching {version} files to {sdir} ")
        # remove the old files, but keep the lunr-pretext-search-index.js file if it exists
        for f in os.listdir(sdir):
            try:
                if "lunr-pretext" not in f and Path(sdir, f).is_file():
                    os.remove(sdir / f)
            except Exception:
                click.echo(f"ERROR - could not delete {sdir} / {f}")
        # call wget non-verbose, recursive, no parents, no hostname, no directoy copy files to sdir
        # trailing slash is important or otherwise you will end up with everything below runestone
        res = subprocess.call(
            f"""wget -nv -r -np -nH -nd -P {sdir} https://runestone.academy/cdn/runestone/{version}/
    """,
            shell=True,
        )
        if res != 0:
            click.echo("wget of runestone files failed")
            return False
    else:
        click.echo(f"_static files already up to date for {version}")
    return True


def manifest_data_to_db(course_name, manifest_path):
    """Read the runestone-manifest.xml file generated by PreTeXt and populate the
    chapters, subchapters, and questions table so that PreTeXt books can be used on
    Runestone.Academy.

    Arguments:
        course_name {string} -- Name of the course (should be a base course)
        manifest_path {path} -- path to runestone-manifest.xml file
    """

    try:
        DBURL = get_dburl()
    except KeyError:
        rslogger.error("PreTeXt integration requires a valid WEB2PY Environment")
        rslogger.error("make sure SERVER_CONFIG and DBURLs are set up")
        exit(-1)

    engine = create_engine(DBURL)
    Session = sessionmaker()
    engine.connect()
    Session.configure(bind=engine)
    sess = Session()
    meta = MetaData()
    chapters = Table("chapters", meta, autoload=True, autoload_with=engine)
    subchapters = Table("sub_chapters", meta, autoload=True, autoload_with=engine)
    questions = Table("questions", meta, autoload=True, autoload_with=engine)
    source_code = Table("source_code", meta, autoload=True, autoload_with=engine)
    course_attributes = Table(
        "course_attributes", meta, autoload=True, autoload_with=engine
    )

    rslogger.info(f"Cleaning up old chapters info for {course_name}")
    # Delete the chapter rows before repopulating. Subchapter rows are taken
    # care of by postgres with the ON DELETE CASCADE clause

    sess.execute(chapters.delete().where(chapters.c.course_id == course_name))

    # Mark existing questions as from_source = 'F' all questions in the manifest will be updated
    # and marked as from_source = 'T' if they are in the manifest.
    sess.execute(
        questions.update()
        .where(questions.c.base_course == course_name)
        .values(from_source="F")
    )

    rslogger.info("Populating the database with Chapter information")
    ext_img_patt = re.compile(r"""src="external""")
    gen_img_patt = re.compile(r"""src="generated""")

    tree = ET.parse(manifest_path)
    root = tree.getroot()
    chap = 0
    for chapter in root.findall("./chapter"):
        rslogger.info(chapter)
        chap += 1
        rslogger.debug(
            f"{chapter.tag} {chapter.find('./id').text} {chapter.find('./title').text}"
        )
        ins = chapters.insert().values(
            chapter_name=chapter.find("./title").text,
            course_id=course_name,
            chapter_label=chapter.find("./id").text,
            chapter_num=chap,
        )
        res = sess.execute(ins)
        chapid = res.inserted_primary_key[0]
        subchap = 0
        #  sub_chapter_name   | character varying(512)
        #  chapter_id         | integer
        #  sub_chapter_label  | character varying(512)
        #  skipreading        | character(1)
        #  sub_chapter_num    | integer
        for subchapter in chapter.findall("./subchapter"):
            subchap += 1

            chap_xmlid = subchapter.find("./id").text
            rslogger.debug(f"subchapter {chap_xmlid}")
            if not chap_xmlid:
                rslogger.error(f"Missing id tag in subchapter {subchapter}")

            titletext = subchapter.find("./title").text
            if not titletext:
                rslogger.debug(f"constructing title for subchapter {chap_xmlid}")
                # ET.tostring  converts the tag and everything to text
                # el.text gets the text inside the element
                titletext = " ".join(
                    [
                        ET.tostring(y).decode("utf8")
                        for y in subchapter.findall("./title/*")
                    ]
                )
            ins = subchapters.insert().values(
                sub_chapter_name=titletext,
                chapter_id=chapid,
                sub_chapter_label=subchapter.find("./id").text,
                skipreading="F",
                sub_chapter_num=subchap,
            )
            sess.execute(ins)

            # Now add this chapter / subchapter to the questions table as a page entry
            name = f"{chapter.find('./title').text}/{subchapter.find('./title').text}"
            res = sess.execute(
                text(
                    """select * from questions where name = :name and base_course = :course_name"""
                ),
                dict(name=name, course_name=course_name),
            ).first()
            valudict = dict(
                base_course=course_name,
                name=name,
                timestamp=datetime.datetime.now(),
                is_private="F",
                question_type="page",
                subchapter=subchapter.find("./id").text,
                chapter=chapter.find("./id").text,
                from_source="T",
            )
            if res:
                ins = (
                    questions.update()
                    .where(
                        and_(
                            questions.c.name == name,
                            questions.c.base_course == course_name,
                        )
                    )
                    .values(**valudict)
                )
            else:
                ins = questions.insert().values(**valudict)
            sess.execute(ins)

            for question in subchapter.findall("./question"):
                dbtext = " ".join(
                    [
                        ET.tostring(y).decode("utf8")
                        for y in question.findall("./htmlsrc/*")
                    ]
                )
                qlabel = " ".join([y.text for y in question.findall("./label")])
                rslogger.debug(f"found label= {qlabel}")
                rslogger.debug("looking for data-component")
                # pdb.set_trace()

                el = question.find(".//*[@data-component]")
                old_ww_id = None
                # Unbelievably if find finds something it evals to False!!
                if el is not None:
                    if "id" in el.attrib:
                        idchild = el.attrib["id"]
                    else:
                        idchild = "fix_me"
                    if "the-id-on-the-webwork" in el.attrib:
                        old_ww_id = el.attrib["the-id-on-the-webwork"]
                else:
                    el = question.find("./div")
                    if el is None:
                        idchild = "fix_me"
                        rslogger.error(
                            f"no id found for question: \n{ET.tostring(question).decode('utf8')}"
                        )
                    elif "id" in el.attrib:
                        idchild = el.attrib["id"]
                    else:
                        idchild = "fix_me"
                try:
                    qtype = el.attrib["data-component"]
                    if qtype == "codelens":
                        # pdb.set_trace()
                        id_el = el.find("./*[@class='pytutorVisualizer']")
                        idchild = id_el.attrib["id"]
                    # translate qtype to question_type
                    qtype = QT_MAP.get(qtype, qtype)
                except Exception:
                    if el is not None:
                        qtype = "webwork"
                        dbtext = ET.tostring(el).decode("utf8")

                if "optional" in question.attrib or qtype == "datafile":
                    optional = "T"
                else:
                    optional = "F"
                practice = "F"
                if qtype == "webwork":
                    practice = "T"
                if el and "practice" in el.attrib:
                    practice = "T"
                autograde = ""
                if "====" in dbtext:
                    extraCode = dbtext.partition("====")[2]  # text after ====
                    # keywords for sql, py, cpp, java respectively
                    for utKeyword in ["assert", "unittest", "TEST_CASE", "junit"]:
                        if utKeyword in extraCode:
                            autograde = "unittest"
                            break
                # chapter and subchapter are elements
                # fix image urls in dbtext to be relative to the book
                dbtext = ext_img_patt.sub(
                    f"""src="/ns/books/published/{course_name}/external""", dbtext
                )
                dbtext = gen_img_patt.sub(
                    f"""src="/ns/books/published/{course_name}/generated""", dbtext
                )
                sbc = subchapter.find("./id").text
                cpt = chapter.find("./id").text
                valudict = dict(
                    base_course=course_name,
                    name=idchild,
                    timestamp=datetime.datetime.now(),
                    is_private="F",
                    question_type=qtype,
                    htmlsrc=dbtext,
                    autograde=autograde,
                    from_source="T",
                    chapter=cpt,
                    subchapter=sbc,
                    topic=f"{cpt}/{sbc}",
                    qnumber=qlabel,
                    optional=optional,
                    practice=practice,
                )
                if old_ww_id:
                    namekey = old_ww_id
                else:
                    namekey = idchild
                res = sess.execute(
                    f"""select * from questions where name='{namekey}' and base_course='{course_name}'"""
                ).first()
                if res:
                    ins = (
                        questions.update()
                        .where(
                            and_(
                                questions.c.name == namekey,
                                questions.c.base_course == course_name,
                            )
                        )
                        .values(**valudict)
                    )
                else:
                    ins = questions.insert().values(**valudict)
                sess.execute(ins)
                if qtype == "datafile":
                    if "data-isimage" in el.attrib:
                        file_contents = el.attrib["src"]
                        if file_contents.startswith("data:"):
                            file_contents = file_contents.split("base64,")[1]
                    else:
                        file_contents = el.text
                    if "data-filename" in el.attrib:
                        filename = el.attrib["data-filename"]
                    else:
                        filename = el.attrib["id"]

                    # write datafile contents to the source_code table
                    res = res = sess.execute(
                        f"""select * from source_code where acid='{filename}' and course_id='{course_name}'"""
                    ).first()

                    vdict = dict(
                        acid=filename, course_id=course_name, main_code=file_contents
                    )
                    if res:
                        upd = (
                            source_code.update()
                            .where(
                                and_(
                                    source_code.c.acid == filename,
                                    source_code.c.course_id == course_name,
                                )
                            )
                            .values(**vdict)
                        )
                    else:
                        upd = source_code.insert().values(**vdict)
                    sess.execute(upd)

    latex = root.find("./latex-macros")
    rslogger.info("Setting attributes for this base course")
    ww_meta = root.find("./webwork-version")
    if ww_meta is not None:
        ww_major = ww_meta.attrib["major"]
        ww_minor = ww_meta.attrib["minor"]
    else:
        ww_major = None
        ww_minor = None

    res = sess.execute(
        f"select * from courses where course_name ='{course_name}'"
    ).first()
    cid = res["id"]

    # Only delete latex_macros and markup_system if they are present. Leave other attributes alone.
    to_delete = ["latex_macros", "markup_system"]
    if ww_major:
        to_delete.append("webwork_js_version")
        to_delete.append("ptx_js_version")
    sess.execute(
        course_attributes.delete().where(
            and_(
                course_attributes.c.course_id == cid,
                course_attributes.c.attr.in_(to_delete),
            )
        )
    )
    if ww_major:
        ins = course_attributes.insert().values(
            course_id=cid, attr="webwork_js_version", value=f"{ww_major}.{ww_minor}"
        )
        ins = course_attributes.insert().values(
            course_id=cid, attr="ptx_js_version", value="0.3"
        )

    ins = course_attributes.insert().values(
        course_id=cid, attr="latex_macros", value=latex.text
    )
    sess.execute(ins)
    ins = course_attributes.insert().values(
        course_id=cid, attr="markup_system", value="PreTeXt"
    )
    sess.execute(ins)
    sess.commit()


class StringIOHandler(logging.Handler):
    """
    A custom logging handler that captures log entries in a StringIO buffer.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stream = StringIO()

    def emit(self, record):
        """
        Emit a record by writing to the StringIO buffer.
        """
        msg = self.format(record)
        self.stream.write(msg + "\n")

    def getvalue(self):
        """
        Return the contents of the StringIO buffer as a string.
        """
        return self.stream.getvalue()

    def flush(self):
        """
        Flush the StringIO buffer.
        """
        self.stream.flush()



# Online Python Tutor
# https://github.com/pgbovine/OnlinePythonTutor/
#
# Copyright (C) Philip J. Guo (philip@pgbovine.net)
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Thanks to John DeNero for making the encoder work on both Python 2 and 3


# Given an arbitrary piece of Python data, encode it in such a manner
# that it can be later encoded into JSON.
#   http://json.org/
#
# We use this function to encode run-time traces of data structures
# to send to the front-end.
#
# Format:
#   Primitives:
#   * None, int, long, float, str, bool - unchanged
#     (json.dumps encodes these fine verbatim, except for inf, -inf, and nan)
#
#   exceptions: float('inf')  -> ['SPECIAL_FLOAT', 'Infinity']
#               float('-inf') -> ['SPECIAL_FLOAT', '-Infinity']
#               float('nan')  -> ['SPECIAL_FLOAT', 'NaN']
#               x == int(x)   -> ['SPECIAL_FLOAT', '%.1f' % x]
#               (this way, 3.0 prints as '3.0' and not as 3, which looks like an int)
#
#   If render_heap_primitives is True, then primitive values are rendered
#   on the heap as ['HEAP_PRIMITIVE', <type name>, <value>]
#
#   (for SPECIAL_FLOAT values, <value> is a list like ['SPECIAL_FLOAT', 'Infinity'])
#
#   Compound objects:
#   * list     - ['LIST', elt1, elt2, elt3, ..., eltN]
#   * tuple    - ['TUPLE', elt1, elt2, elt3, ..., eltN]
#   * set      - ['SET', elt1, elt2, elt3, ..., eltN]
#   * dict     - ['DICT', [key1, value1], [key2, value2], ..., [keyN, valueN]]
#   * instance - ['INSTANCE', class name, [attr1, value1], [attr2, value2], ..., [attrN, valueN]]
#   * instance with __str__ defined - ['INSTANCE_PPRINT', class name, <__str__ value>]
#   * class    - ['CLASS', class name, [list of superclass names], [attr1, value1], [attr2, value2], ..., [attrN, valueN]]
#   * function - ['FUNCTION', function name, parent frame ID (for nested functions)]
#   * module   - ['module', module name]
#   * other    - [<type name>, string representation of object]
#   * compound object reference - ['REF', target object's unique_id]
#
# the unique_id is derived from id(), which allows us to capture aliasing


# number of significant digits for floats
FLOAT_PRECISION = 4


from collections import defaultdict
import re, types
import sys
import math

typeRE = re.compile("<type '(.*)'>")
classRE = re.compile("<class '(.*)'>")

import inspect

# TODO: maybe use the 'six' library to smooth over Py2 and Py3 incompatibilities?
is_python3 = sys.version_info[0] == 3
if is_python3:
    # avoid name errors (GROSS!)
    long = int
    unicode = str


def is_class(dat):
    """Return whether dat is a class."""
    if is_python3:
        return isinstance(dat, type)
    else:
        return type(dat) in (types.ClassType, types.TypeType)


def is_instance(dat):
    """Return whether dat is an instance of a class."""
    if is_python3:
        return (
            type(dat) not in PRIMITIVE_TYPES
            and isinstance(type(dat), type)
            and not isinstance(dat, type)
        )
    else:
        # ugh, classRE match is a bit of a hack :(
        return type(dat) == types.InstanceType or classRE.match(str(type(dat)))


def get_name(obj):
    """Return the name of an object."""
    return obj.__name__ if hasattr(obj, "__name__") else get_name(type(obj))


PRIMITIVE_TYPES = (int, long, float, str, unicode, bool, type(None))


def encode_primitive(dat):
    t = type(dat)
    if t is float:
        if math.isinf(dat):
            if dat > 0:
                return ["SPECIAL_FLOAT", "Infinity"]
            else:
                return ["SPECIAL_FLOAT", "-Infinity"]
        elif math.isnan(dat):
            return ["SPECIAL_FLOAT", "NaN"]
        else:
            # render floats like 3.0 as '3.0' and not as 3
            if dat == int(dat):
                return ["SPECIAL_FLOAT", "%.1f" % dat]
            else:
                return round(dat, FLOAT_PRECISION)
    elif t is str and (not is_python3):
        # hack only for Python 2 strings ... always turn into unicode
        # and display '?' when it's not valid unicode
        return dat.decode("utf-8", "replace")
    else:
        # return all other primitives verbatim
        return dat


# grab a line number like ' <line 2>' or ' <line 2b>'
def create_lambda_line_number(codeobj, line_to_lambda_code):
    try:
        lambda_lineno = codeobj.co_firstlineno
        lst = line_to_lambda_code[lambda_lineno]
        ind = lst.index(codeobj)
        # add a suffix for all subsequent lambdas on a line beyond the first
        # (nix this for now because order isn't guaranteed when you have
        #  multiple lambdas on the same line)
        """
    if ind > 0:
      lineno_str = str(lambda_lineno) + chr(ord('a') + ind)
    else:
      lineno_str = str(lambda_lineno)
    """
        lineno_str = str(lambda_lineno)
        return " <line " + lineno_str + ">"
    except:
        return ""


# Note that this might BLOAT MEMORY CONSUMPTION since we're holding on
# to every reference ever created by the program without ever releasing
# anything!
class ObjectEncoder:
    def __init__(self, render_heap_primitives):
        # Key: canonicalized small ID
        # Value: encoded (compound) heap object
        self.encoded_heap_objects = {}

        self.render_heap_primitives = render_heap_primitives

        self.id_to_small_IDs = {}
        self.cur_small_ID = 1

        # wow, creating unique identifiers for lambdas is quite annoying,
        # especially if we want to properly differentiate:
        # 1.) multiple lambdas defined on the same line, and
        # 2.) the same lambda code defined multiple times on different lines
        #
        # However, it gets confused when there are multiple identical
        # lambdas on the same line, like:
        # f(lambda x:x*x, lambda y:y*y, lambda x:x*x)

        # (assumes everything is in one file)
        # Key:   line number
        # Value: list of the code objects of lambdas defined
        #        on that line in the order they were defined
        self.line_to_lambda_code = defaultdict(list)

    def get_heap(self):
        return self.encoded_heap_objects

    def reset_heap(self):
        # VERY IMPORTANT to reassign to an empty dict rather than just
        # clearing the existing dict, since get_heap() could have been
        # called earlier to return a reference to a previous heap state
        self.encoded_heap_objects = {}

    def set_function_parent_frame_ID(self, ref_obj, enclosing_frame_id):
        assert ref_obj[0] == "REF"
        func_obj = self.encoded_heap_objects[ref_obj[1]]
        assert func_obj[0] == "FUNCTION"
        func_obj[-1] = enclosing_frame_id

    # return either a primitive object or an object reference;
    # and as a side effect, update encoded_heap_objects
    def encode(self, dat, get_parent):
        """Encode a data value DAT using the GET_PARENT function for parent ids."""
        # primitive type
        if not self.render_heap_primitives and type(dat) in PRIMITIVE_TYPES:
            return encode_primitive(dat)
        # compound type - return an object reference and update encoded_heap_objects
        else:
            my_id = id(dat)

            try:
                my_small_id = self.id_to_small_IDs[my_id]
            except KeyError:
                my_small_id = self.cur_small_ID
                self.id_to_small_IDs[my_id] = self.cur_small_ID
                self.cur_small_ID += 1

            del my_id  # to prevent bugs later in this function

            ret = ["REF", my_small_id]

            # punt early if you've already encoded this object
            if my_small_id in self.encoded_heap_objects:
                return ret

            # major side-effect!
            new_obj = []
            self.encoded_heap_objects[my_small_id] = new_obj

            typ = type(dat)

            if typ == list:
                new_obj.append("LIST")
                for e in dat:
                    new_obj.append(self.encode(e, get_parent))
            elif typ == tuple:
                new_obj.append("TUPLE")
                for e in dat:
                    new_obj.append(self.encode(e, get_parent))
            elif typ == set:
                new_obj.append("SET")
                for e in dat:
                    new_obj.append(self.encode(e, get_parent))
            elif typ == dict:
                new_obj.append("DICT")
                for (k, v) in dat.items():
                    # don't display some built-in locals ...
                    if k not in ("__module__", "__return__", "__locals__"):
                        new_obj.append(
                            [self.encode(k, get_parent), self.encode(v, get_parent)]
                        )
            elif typ in (types.FunctionType, types.MethodType):
                if is_python3:
                    argspec = inspect.getfullargspec(dat)
                else:
                    argspec = inspect.getargspec(dat)

                printed_args = [e for e in argspec.args]
                if argspec.varargs:
                    printed_args.append("*" + argspec.varargs)

                if is_python3:
                    if argspec.varkw:
                        printed_args.append("**" + argspec.varkw)
                    if argspec.kwonlyargs:
                        printed_args.extend(argspec.kwonlyargs)
                else:
                    if argspec.keywords:
                        printed_args.append("**" + argspec.keywords)

                func_name = get_name(dat)

                pretty_name = func_name

                # sometimes might fail for, say, <genexpr>, so just ignore
                # failures for now ...
                try:
                    pretty_name += "(" + ", ".join(printed_args) + ")"
                except TypeError:
                    pass

                # put a line number suffix on lambdas to more uniquely identify
                # them, since they don't have names
                if func_name == "<lambda>":
                    cod = dat.__code__ if is_python3 else dat.func_code  # ugh!
                    lst = self.line_to_lambda_code[cod.co_firstlineno]
                    if cod not in lst:
                        lst.append(cod)
                    pretty_name += create_lambda_line_number(
                        cod, self.line_to_lambda_code
                    )

                encoded_val = ["FUNCTION", pretty_name, None]
                if get_parent:
                    enclosing_frame_id = get_parent(dat)
                    encoded_val[2] = enclosing_frame_id
                new_obj.extend(encoded_val)
            elif typ is types.BuiltinFunctionType:
                pretty_name = get_name(dat) + "(...)"
                new_obj.extend(["FUNCTION", pretty_name, None])
            elif is_class(dat) or is_instance(dat):
                self.encode_class_or_instance(dat, new_obj)
            elif typ is types.ModuleType:
                new_obj.extend(["module", dat.__name__])
            elif typ in PRIMITIVE_TYPES:
                assert self.render_heap_primitives
                new_obj.extend(
                    ["HEAP_PRIMITIVE", type(dat).__name__, encode_primitive(dat)]
                )
            else:
                typeStr = str(typ)
                m = typeRE.match(typeStr)

                if not m:
                    m = classRE.match(typeStr)

                assert m, typ

                if is_python3:
                    encoded_dat = str(dat)
                else:
                    # ugh, for bytearray() in Python 2, str() returns
                    # non-JSON-serializable characters, so need to decode:
                    encoded_dat = str(dat).decode("utf-8", "replace")
                new_obj.extend([m.group(1), encoded_dat])

            return ret

    def encode_class_or_instance(self, dat, new_obj):
        """Encode dat as a class or instance."""
        if is_instance(dat):
            if hasattr(dat, "__class__"):
                # common case ...
                class_name = get_name(dat.__class__)
            else:
                # super special case for something like
                # "from datetime import datetime_CAPI" in Python 3.2,
                # which is some weird 'PyCapsule' type ...
                # http://docs.python.org/release/3.1.5/c-api/capsule.html
                class_name = get_name(type(dat))

            if hasattr(dat, "__str__") and (
                not dat.__class__.__str__ is object.__str__
            ):  # make sure it's not the lame default __str__
                # N.B.: when objects are being constructed, this call
                # might fail since not all fields have yet been populated
                try:
                    pprint_str = str(dat)
                except:
                    pprint_str = "<incomplete object>"

                new_obj.extend(["INSTANCE_PPRINT", class_name, pprint_str])
                return  # bail early
            else:
                new_obj.extend(["INSTANCE", class_name])
                # don't traverse inside modules, or else risk EXPLODING the visualization
                if class_name == "module":
                    return
        else:
            superclass_names = [e.__name__ for e in dat.__bases__ if e is not object]
            new_obj.extend(["CLASS", get_name(dat), superclass_names])

        # traverse inside of its __dict__ to grab attributes
        # (filter out useless-seeming ones, based on anecdotal observation):
        hidden = (
            "__doc__",
            "__module__",
            "__return__",
            "__dict__",
            "__locals__",
            "__weakref__",
            "__qualname__",
        )
        if hasattr(dat, "__dict__"):
            user_attrs = sorted([e for e in dat.__dict__ if e not in hidden])
        else:
            user_attrs = []

        for attr in user_attrs:
            new_obj.append(
                [self.encode(attr, None), self.encode(dat.__dict__[attr], None)]
            )



# Copyright (C) 2012  Michael Hewner, Isaiah Mayerchak
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "isaiahmayerchak"
from textwrap import dedent
import pdb
import random
from docutils import nodes
from docutils.parsers.rst import directives
from runestone.mchoice import Assessment
from runestone.server.componentdb import (
    addQuestionToDB,
    addHTMLToDB,
    maybeAddToAssignment,
)
from runestone.common.runestonedirective import RunestoneIdNode


def setup(app):
    app.add_directive("parsonsprob", ParsonsProblem)
    app.add_node(
        ParsonsNode,
        html=(visit_parsons_html, depart_parsons_html),
        xml=(visit_parsons_xml, depart_parsons_xml),
    )
    app.add_config_value("parsons_div_class", "runestone", "html")


TEMPLATE_START = """
        <div class="%(divclass)s parsons-container %(optclass)s">
        <div data-component="parsons" id="%(divid)s" class="parsons" >
        <div class="parsons_question parsons-text" >
    """

TEMPLATE_END = """
        </div>
        <pre  class="parsonsblocks" data-question_label="%(question_label)s"  %(adaptive)s %(maxdist)s %(order)s %(noindent)s %(language)s %(grader)s %(numbered)s %(optional)s style="visibility: hidden;">
        %(code)s
        </pre>
        </div>
        </div>
    """


class ParsonsNode(nodes.General, nodes.Element, RunestoneIdNode):
    pass


def wrap_with_tag(text: str, tag: str, **kwargs) -> str:
    start = f"<{tag}"
    for k, v in kwargs.items():
        start += f" {k}='{v}'"
    start += ">"
    end = f"</{tag}>\n"
    return start + text + end


def visit_parsons_xml(self, node):
    if node["runestone_options"]["numbered"]:
        node["runestone_options"]["numbered"] = "numbered='yes'"

    if node["runestone_options"]["adaptive"]:
        node["runestone_options"]["adaptive"] = "adaptive='yes'"

    if node["runestone_options"]["noindent"]:
        node["runestone_options"]["noindent"] = "indentation='show'"
    else:
        node["runestone_options"]["noindent"] = "indentation='hide'"

    if node["runestone_options"]["language"]:
        node["runestone_options"]["language"] = node["runestone_options"][
            "language"
        ].replace("data-", "")
    else:
        node["runestone_options"]["language"] = "language='python'"

    res = (
        "<exercise label='{divid}' {numbered} {adaptive} {noindent} {language}>".format(
            **node["runestone_options"]
        )
    )
    res += "<statement>\n"
    self.output.append(res)


def process_one_block(block, choice=False, correct=False, natural=False):
    code_list = block.split("\n")
    if choice:
        attrs = "correct='yes'" if correct else ""
        res = f"<choice {attrs}>"
    else:
        res = ""
    for line in code_list:
        if line:
            line = line.replace("#paired", "")
            line = line.replace("#distractor", "")
            if natural:
                res += wrap_with_tag(line, "p")
            else:
                res += wrap_with_tag(line, "cline")
    if choice:
        res += "</choice>"
    return res


def depart_parsons_xml(self, node):
    res = "</statement>\n<blocks>"
    if "natural" in node["runestone_options"]["language"]:
        natural = True
    else:
        natural = False
    # first split into blocks based on --- then process each block...
    block_list = dedent(node["runestone_options"]["code"]).split("---")
    order = list(range(1, len(block_list) + 1))
    random.shuffle(order)
    block = block_list.pop(0)
    while block:
        attr = f"order='{order.pop()}' "
        attr += "correct='no'" if "#distractor" in block else ""
        res += f"<block {attr}>\n"
        # check to see if the next block is a distractor
        if block_list and "#paired" in block_list[0]:
            choice = True
            res += process_one_block(block, choice, True, natural)
            res += process_one_block(block_list.pop(0), choice, False, natural)
        else:
            choice = False
            res += process_one_block(block, choice, True, natural)
        res += "</block>"
        if block_list:
            block = block_list.pop(0)
        else:
            block = None

    res += "</blocks></exercise>"
    self.output.append(res)


def visit_parsons_html(self, node):
    node["delimiter"] = "_start__{}_".format(node["runestone_options"]["divid"])
    self.body.append(node["delimiter"])
    res = TEMPLATE_START % node["runestone_options"]
    self.body.append(res)


def depart_parsons_html(self, node):
    res = TEMPLATE_END % node["runestone_options"]
    self.body.append(res)
    addHTMLToDB(
        node["runestone_options"]["divid"],
        node["runestone_options"]["basecourse"],
        "".join(self.body[self.body.index(node["delimiter"]) + 1 :]),
    )
    self.body.remove(node["delimiter"])


class ParsonsProblem(Assessment):
    """
    .. parsonsprob:: unqiue_problem_id_here
       :maxdist:
       :order:
       :language:
       :noindent:
       :adaptive:
       :numbered:

       Solve my really cool parsons problem...if you can.
       -----
       def findmax(alist):
       =====
          if len(alist) == 0:
             return None
       =====
          curmax = alist[0]
          for item in alist:
       =====
             if item &gt; curmax:
       =====
                curmax = item
       =====
          return curmax


    config values (conf.py):

    - parsons_div_class - custom CSS class of the component's outermost div
    """

    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = False
    option_spec = Assessment.option_spec.copy()
    option_spec.update(
        {
            "maxdist": directives.unchanged,
            "order": directives.unchanged,
            "language": directives.unchanged,
            "noindent": directives.flag,
            "adaptive": directives.flag,
            "numbered": directives.unchanged,
            "grader": directives.unchanged,
        }
    )
    has_content = True

    def run(self):
        """

           Instructions for solving the problem should be written and then a line with -----
           signals the beginning of the code.  If you want more than one line in a single
           code block, seperate your code blocks with =====.

           Both the instructions sections and code blocks are optional. If you don't include any
           =====, the code will assume you want each line to be its own code block.

        Example:

        .. parsonsprob:: unqiue_problem_id_here

           Solve my really cool parsons problem...if you can.
           -----
           def findmax(alist):
           =====
              if len(alist) == 0:
                 return None
           =====
              curmax = alist[0]
              for item in alist:
           =====
                 if item &gt; curmax:
           =====
                    curmax = item
           =====
              return curmax


        """

        super(ParsonsProblem, self).run()
        addQuestionToDB(self)

        env = self.state.document.settings.env
        self.options["instructions"] = ""
        self.options["code"] = self.content
        self.options["divclass"] = env.config.parsons_div_class

        if "numbered" in self.options:
            self.options["numbered"] = (
                ' data-numbered="' + self.options["numbered"] + '"'
            )  # ' data-numbered="true"'
        else:
            self.options["numbered"] = ""

        if "maxdist" in self.options:
            self.options["maxdist"] = ' data-maxdist="' + self.options["maxdist"] + '"'
        else:
            self.options["maxdist"] = ""
        if "order" in self.options:
            self.options["order"] = ' data-order="' + self.options["order"] + '"'
        else:
            self.options["order"] = ""
        if "noindent" in self.options:
            self.options["noindent"] = ' data-noindent="true"'
        else:
            self.options["noindent"] = ""
        if "adaptive" in self.options:
            if self.options.get("grader") == "dag" and "adaptive" in self.options:
                raise Exception("Adaptivity not yet supported with DAG grader")
            self.options["adaptive"] = ' data-adaptive="true"'
        else:
            self.options["adaptive"] = ""
        if "language" in self.options:
            self.options["language"] = (
                ' data-language="' + self.options["language"] + '"'
            )
        else:
            self.options["language"] = ""
        if "grader" in self.options:
            self.options["grader"] = ' data-grader="' + self.options["grader"] + '"'
        else:
            self.options["grader"] = ""

        if "-----" in self.content:
            index = self.content.index("-----")
            self.options["instructions"] = self.content[:index]
            self.options["code"] = self.content[index + 1 :]
        else:
            self.options["instructions"] = ["Arrange the blocks"]

        if "=====" in self.options["code"]:
            self.options["code"] = "\n".join(self.options["code"])

            self.options["code"] = self.options["code"].replace("=====", "---")
        else:
            self.options["code"] = "\n".join(self.options["code"])

        self.assert_has_content()

        maybeAddToAssignment(self)
        parsons_node = ParsonsNode()
        parsons_node["runestone_options"] = self.options
        (
            parsons_node["source"],
            parsons_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)
        self.state.nested_parse(
            self.options["instructions"], self.content_offset, parsons_node
        )
        # explain_text is a list.
        return [parsons_node]



# Copyright (C) 2017 Wayne Brown
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import print_function

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneNode,
)
import re

__author__ = "Wayne Brown"

# This creates a matrix equation
#
# HTML and CSS objects for the design:
# <div class="matrixeq">
#   <span class="matrix"> .. </span>
#   <span class="matrix_operator"> .. </span>
#   <span class="matrix"> .. </span>
#   ...
#   <span class="matrix"> .. </span>
#   <span class="equation_number"> .. </span>
# </div>
#


# --------------------------------------------------------------------------
def setup(app):
    app.add_directive("matrixeq", MatrixEq)
    app.add_role("inline_matrixeq", inline_matrixeq)

    app.add_autoversioned_stylesheet("matrixeq.css")

    app.add_autoversioned_javascript("matrixeq.js", defer="")

    app.add_node(MatrixEqNode, html=(visit_matrixeq_html, depart_matrixeq_html))
    app.add_node(
        InlineMatrixEqNode,
        html=(visit_inline_matrixeq_html, depart_inline_matrixeq_html),
    )

    app.connect("doctree-resolved", process_matrixeq_nodes)
    app.connect("env-purge-doc", purge_matrixeq)


# ==========================================================================
# A python class derived from the sphinx class Directive
# This initializes the new directive
class MatrixEq(RunestoneDirective):
    """
    .. matrixeq:: uniqueid
        :notexecutable: -- the matrix equation can't be executed by the user
        :comment: -- A comment to include to the right of the equation
        :nolabel: -- don't label the equation using the uniqueid provided
        :backgroundcolor: -- the color of the background; either #RRBBGG or a color name
        :foregroundcolor: -- the color used for the matrix elements; either #RRBBGG or a color name
        :highlightcolor: -- the color used for "bolded" elements; ; either #RRBBGG or a color name

        A single matrix is defined using javascript array notation, e.g., [a,b,c;d,e,f]

        A "name" can be assigned to a matrix by including a string and a colon after the
        beginning [, but before the first value, e.g., [M1: a,b,c;d,e,f]. If no name is
        specified, a default name is assigned.

        A background color can be assigned to a matrix by including a color specifier
        after the matrix's name. The color can be one of the standard HTML color names
        or a color value (#RRGGBB). To specify a color, you must give the matrix a name.
        E.g., [M1,lightcyan: a,b,c;d,e,f] or [M1,#DF85E8: a,b,c;d,e,f]

        Individual elements of a matrix can have no special formatting, be highlighted,
        be editable, or be both highlighted and editable. To make an element:
            - highlighted, add an asterisk in front of the value, e.g., *a
            - editable, put the value inside {}, e.g, {a}
            - both highlighted and editable, *{a} or {*a}

        The content of the directive is a matrix equation. E.g.,
        [M1: 1, 0, 0, 0; 0, 1, 0, 0; 0, 0, -c2, c1; 0, 0, -1, 0]*[M2: x;y;z;1] = [M3: x';y';z';w']
    """

    required_arguments = 1
    optional_arguments = 0
    has_content = True
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update(
        {
            "notexecutable": directives.flag,
            "comment": directives.unchanged,
            "nolabel": directives.flag,
            "backgroundcolor": directives.unchanged,
            "foregroundcolor": directives.unchanged,
            "highlightcolor": directives.unchanged,
        }
    )

    def run(self):

        # print("self.options = ", self.options)

        env = self.state.document.settings.env
        if not hasattr(env, "matrixeqcounter"):
            env.matrixeqcounter = 0
        env.matrixeqcounter += 1

        # The contents is a StringList. Combine all the strings into a single string
        # and put a newline character at the line breaks.
        text = "".join(self.content)

        text = text.replace("u'", "'")  # hack: avoid unicode strings

        self.options["name"] = self.arguments[0].strip()

        self.options["contents"] = text

        self.options["equationnumber"] = self.arguments[0]
        self.options["equationcounter"] = env.matrixeqcounter
        self.options["executable"] = "notexecutable" not in self.options
        self.options["nolabel"] = "nolabel" in self.options
        if "comment" in self.options:
            self.options["comment"] = self.options["comment"].strip()
        else:
            self.options["comment"] = ""

        color_scheme = ' style="background-color:'
        if "backgroundcolor" in self.options:
            color_scheme += self.options["backgroundcolor"].strip() + ";"
        else:
            color_scheme += "#fcf8e3;"

        if "foregroundcolor" in self.options:
            color_scheme += " color:" + self.options["foregroundcolor"].strip() + ";"

        self.options["colorscheme"] = color_scheme + '"'

        if "highlightcolor" in self.options:
            self.options["highlightcolor"] = self.options["highlightcolor"].strip()
        else:
            self.options["highlightcolor"] = "red"  # default highlight color
        meqn = MatrixEqNode()
        meqn["components"] = self.options
        return [meqn]


# ==========================================================================
class MatrixEqNode(nodes.General, nodes.Element, RunestoneNode):
    pass


def matrixToHTML(text, nodeID, node):
    # Divide the text into a "header" and the matrix values
    parts = text.split(":")
    if len(parts) == 1:
        header = nodeID  # The generated ID from the calling function
        valuesStr = parts[0]  # The entire text
    else:
        header = parts[0]  # All text before the :
        valuesStr = parts[1]  # All text after the :

    # Check to see if there is a color value after the node name string
    if "," in header:
        nodeName, matrixColor = header.split(",")
        matrixColor = ' style="background-color: ' + matrixColor + '" '
    else:
        nodeName = header
        matrixColor = ""

    if "*" in nodeName:
        highlight = ' style="color:#C8255D" '
    else:
        highlight = ""

    # formatting codes
    # 0: no formatting
    # 1: make the element highlighted
    # 2: make the element editable
    # 3: make the element both editable and highlighted

    # Define the output format. Make this user controllable in the future
    fieldSize = 6
    precision = str(fieldSize) + "." + str(3) + "f"

    # Put the matrix values into a 2D array
    values = []
    valuesFormat = []

    rowStrings = valuesStr.split(";")
    for row in range(len(rowStrings)):

        rowTextValues = rowStrings[row].split(",")

        rowValues = []
        rowFormat = []
        for col in range(len(rowTextValues)):
            # remove all white space around the value
            s = rowTextValues[col].strip()

            formatValue = 0
            # print( "value = " + rowTextValues[col])
            if s[0] == "*":
                formatValue |= 1
                s = s[1:]  # remove the formatting symbol

            if s[0] == "{":
                formatValue |= 2
                s = s[1:]  # remove the formatting symbol
                if s[-1] == "}":
                    s = s[:-1]

            if s[0] == "*":
                formatValue |= 1
                s = s[1:]  # remove the formatting symbol

            # print( "value = " + s)
            try:
                valueAsNumber = float(s)
                if valueAsNumber.is_integer():
                    valueAsString = format(valueAsNumber, "6d")
                else:
                    valueAsString = format(valueAsNumber, precision)
            except ValueError:
                valueAsString = s

            rowValues.append(valueAsString)
            rowFormat.append(formatValue)

        values.append(rowValues)
        valuesFormat.append(rowFormat)

    # print("values = ", values);
    # print("formats = ", valuesFormat);

    # Build the HTML for the matrix
    res = (
        '<span id="' + nodeID + '" class="matrix_table"' + highlight + matrixColor + ">"
    )

    nRows = len(values)
    nCols = len(values[0])
    # Check to make sure every row has the same number of columns
    for r in range(0, nRows):
        if len(values[r]) != nCols:
            res = (
                "matrixeq directive error: row "
                + str(r)
                + " does not have "
                + str(nCols)
                + " values <br />"
            )
            res += text
            return res, nRows

    # print("matrix size = ", nRows, nCols)

    # Create the HTML code for the matrix
    for c in range(0, nCols):
        res += '<span class="matrix_column">'  # start column

        for r in range(0, nRows):
            valueStr = values[r][c]

            # Replace any exponential's with html <sup>v</sup>
            expIndex = valueStr.find("^(")
            if expIndex >= 0:
                expIndexEnd = valueStr.find(")", expIndex)
                if expIndexEnd > 0:
                    exp = valueStr[expIndex + 2 : expIndexEnd]
                    valueStr = (
                        valueStr[0:expIndex]
                        + "<sup>"
                        + exp
                        + "</sup>"
                        + valueStr[expIndexEnd + 1 :]
                    )

            # Create the correct type of field
            if valuesFormat[r][c] == 0:
                res += "<span>" + valueStr + "<br /></span>"
            elif valuesFormat[r][c] == 1:
                res += (
                    '<span style="color:'
                    + node["components"]["highlightcolor"]
                    + ';">'
                    + valueStr
                    + "<br /></span>"
                )
            elif valuesFormat[r][c] == 2:
                res += '<span><input type="text" value="' + valueStr + '"></span>'
            elif valuesFormat[r][c] == 3:
                res += (
                    '<span><input type="text" value="'
                    + valueStr
                    + '" style="color:'
                    + node["components"]["highlightcolor"]
                    + '";></span>'
                )

        res += "</span>"  # end column

    res += "</span>"  # END_MATRIX

    return res, nRows


# parse a matrix equation into its distinct parts
def divide_matrixeq_into_its_parts(text):
    parts = []
    # print("text = ", text)
    while len(text) > 0:
        matrix_start = text.find("[")
        # print("matrix_start = ", matrix_start)
        if matrix_start >= 0:
            matrix_end = text.find("]", matrix_start)
            before = text[0:matrix_start]
            matrix = text[matrix_start : matrix_end + 1]
            text = text[matrix_end + 1 :]
        else:
            before = text
            matrix = ""
            text = ""

        before = before.strip()
        if len(before) > 0:
            parts.append(before)
        if len(matrix) > 0:
            parts.append(matrix)

    # print("matrixeq parts: ", parts)
    return parts


# self for these functions is an instance of the writer class.  For example
# in html, self is sphinx.writers.html.SmartyPantsHTMLTranslator
# The node that is passed as a parameter is an instance of our node class.
def visit_matrixeq_html(self, node):
    # print("In visit_matrixeq_html, node["components"] = ", node["components"])

    # Parse the matrix equation into its parts
    parts = divide_matrixeq_into_its_parts(node["components"]["contents"].strip())

    id = "M" + str(node["components"]["equationcounter"])
    node["components"]["equationcounter"] += 1

    # start of HTML
    res = "<!-- matrixeq start -->\n"
    res += (
        "<div id='"
        + id
        + "' class='matrixeq_container'"
        + node["components"]["colorscheme"]
        + ">\n"
    )

    for j in range(0, len(parts)):
        if parts[j][0] == "[":
            (text, nRows) = matrixToHTML(parts[j][1:-1], id + "_" + str(j), node)
            res += text
        else:
            if node["components"]["executable"]:
                event = ' onclick="Matrixeq_directive(this);"'
            else:
                event = ""
            res += '<span class="matrix_operator"' + event + ">" + parts[j] + "</span>"

    # Add a comment to the end of the equation
    comment = ""
    if len(node["components"]["comment"]) > 0:
        comment = " - " + node["components"]["comment"]

    label = node["components"]["equationnumber"]
    if node["components"]["nolabel"]:
        label = ""

    res += "<span class='matrix_label'> " + label + comment + "</span>"

    # End of HTML
    res += "</div>\n"
    res += "<!-- end of matrixeq -->"

    self.body.append(res)


# --------------------------------------------------------------------------
def depart_matrixeq_html(self, node):
    """
    This is called at the start of processing an activecode node.  If activecode had recursive nodes
    etc and did not want to do all of the processing in visit_matrixeq_html any finishing touches could be
    added here.
    """
    pass


# --------------------------------------------------------------------------
def process_matrixeq_nodes(app, env, docname):
    pass


# --------------------------------------------------------------------------
def purge_matrixeq(app, env, docname):
    pass


# ==========================================================================
# Inline_MatrixEq role
# ==========================================================================
# Create a role representing a matrix equation that can be included into an
# in-line paragraph.
class InlineMatrixEqNode(nodes.General, nodes.Element, RunestoneNode):
    """
    :inline_matrixeq:`[a,b;c,d]`

        A inline_matrixeq role allows a matrix equation to be defined inside a
        paragraph. The syntax for the matrix equation is identical to a matrixeq
        directive. The operators in an in-line matrix equation are not executable.

        In the future it would be nice to figure out how to make the background
        color user configurable. (For some reason, inheriting the background color
        from the enclosing parent makes the brackets of the matrices render
        incorrectly.)

        The background color is hardcoded to a light yellow.
        The highlight color is hardcoded to red.
    """

    pass


def inline_matrixeq(
    roleName,  # _`roleName`: the local name of the interpreted role, the role name actually used in
    # the document.
    rawtext,  # _`rawtext` is a string containing the entire interpreted text input, including the
    # role and markup. Return it as a problematic node linked to a system message if a
    # problem is encountered.
    text,  # The interpreted _`text` content.
    lineno,  # The line number (_`lineno`) where the interpreted text begins.
    # _`inliner` is the docutils.parsers.rst.states.Inliner object that called this function.
    inliner,
    # It contains the several attributes useful for error reporting and document tree access.
    # A dictionary of directive _`options` for customization (from the "role" directive),
    options={},
    # to be interpreted by this function. Used for additional attributes for the generated elements and other functionality.
    content=[],
):  # A list of strings, the directive _`content` for customization (from the "role"
    # directive). To be interpreted by the role function.
    """ """
    matrix_node = InlineMatrixEqNode()
    matrix_text = re.search(":inline_matrixeq:`(.*)`", rawtext).group(1)
    matrix_node["components"] = {
        "contents": matrix_text,
        "colorscheme": ' style="background-color:inherit; color: inherit"',
        "highlightcolor": "red",
        "equationcounter": 0,
    }

    matrix_node["line"] = lineno
    return [matrix_node], []


def visit_inline_matrixeq_html(self, node):

    # Parse the matrix equation into its parts
    parts = divide_matrixeq_into_its_parts(node["components"]["contents"].strip())

    id = "M" + str(node["components"]["equationcounter"])
    node["components"]["equationcounter"] += 1

    # start of HTML
    res = "<!-- inline_matrixeq start -->\n"
    res += (
        "<span id='"
        + id
        + "' class='matrixeq_container'"
        + node["components"]["colorscheme"]
        + ">\n"
    )

    for j in range(0, len(parts)):
        if parts[j][0] == "[":
            (text, nRows) = matrixToHTML(parts[j][1:-1], id + "_" + str(j), node)
            res += text
        else:
            res += '<span class="matrix_operator">' + parts[j] + "</span>"

    # End of HTML
    res += "</span>\n"
    res += "<!-- end of inline_matrixeq -->"

    self.body.append(res)


def depart_inline_matrixeq_html(self, node):
    pass



# **************************************
# |docname| - A for loop for Sphinx/reST
# **************************************
# This provides a simple for loop for Sphinx/reST.
#
# Imports
# =======
# These are listed in the order prescribed by `PEP 8
# <http://www.python.org/dev/peps/pep-0008/#imports>`_.
#
# Standard library
# ----------------
# None.
#
# Third-party imports
# -------------------
from docutils.parsers.rst import directives
from docutils.parsers.rst import Directive
from docutils import nodes
from docutils.statemachine import StringList

# Local application imports
# -------------------------
# None.


class ForLoop(Directive):
    # The required argument is the first parameter to ``range``.
    required_arguments = 1
    # The optional arguments are the second and third parameter to ``range``.
    optional_arguments = 2
    # Per http://docutils.sourceforge.net/docs/howto/rst-directives.html, True if content is allowed. However, this isn't checked or enforced.
    has_content = True

    def __init__(self, *args, **kwargs):
        super(ForLoop, self).__init__(*args, **kwargs)

    def run(self):
        # Create and error-check the range.
        try:
            _range = range(*[int(arg) for arg in self.arguments])
        except Exception as e:
            raise self.error("Invalid arguments to range: {}".format(e))

        # Run the for loop over all this directive's content. Docutils expects a StringList, so use that.
        loop_content = StringList()
        for loop_index in _range:
            # Loop over each line of the content, only replacing lines in the content of this directive.
            for source, offset, value in self.content.xitems():
                loop_content.append(
                    value.format(loop_index, *self.arguments), source, offset
                )

            # Add an additional newline between loop iterations.
            loop_content.append("\n", "for-loop", 0)

        # Parse the resulting content and return it.
        node = nodes.container()
        self.state.nested_parse(loop_content, self.content_offset, node)
        return [node]


directives.register_directive("for-loop", ForLoop)



# Copyright (C) 2011  Bradley N. Miller
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = "isaiahmayerchak"

from docutils import nodes
from docutils.parsers.rst import directives
from runestone.common.runestonedirective import (
    RunestoneDirective,
    RunestoneNode,
)


# add directives/javascript/css
def setup(app):
    app.add_directive("reveal", RevealDirective)

    app.add_node(RevealNode, html=(visit_reveal_html, depart_reveal_html))


class RevealNode(nodes.General, nodes.Element, RunestoneNode):
    pass


def visit_reveal_html(self, node):
    # Set options and format templates accordingly

    if "modal" in node["runestone_options"]:
        node["runestone_options"]["modal"] = "data-modal"
    else:
        node["runestone_options"]["modal"] = ""

    if "modaltitle" in node["runestone_options"]:
        temp = node["runestone_options"]["modaltitle"]
        node["runestone_options"]["modaltitle"] = """data-title=""" + '"' + temp + '"'
    else:
        node["runestone_options"]["modaltitle"] = ""

    if (
        node["runestone_options"]["instructoronly"]
        and node["runestone_options"]["is_dynamic"]
    ):
        res = DYNAMIC_PREFIX
    else:
        res = ""

    res += TEMPLATE_START % node["runestone_options"]
    self.body.append(res)


def depart_reveal_html(self, node):
    # Set options and format templates accordingly
    res = TEMPLATE_END % node["runestone_options"]
    if (
        node["runestone_options"]["instructoronly"]
        and node["runestone_options"]["is_dynamic"]
    ):
        res += DYNAMIC_SUFFIX

    self.body.append(res)


# Templates to be formatted by node options
DYNAMIC_PREFIX = """
{% if is_instructor: %}
"""
TEMPLATE_START = """
    <div data-component="reveal" class="runestone %(optclass)s" id="%(divid)s" %(modal)s %(modaltitle)s %(showtitle)s %(hidetitle)s %(instructoronly)s style="visibility: hidden;">
    """
TEMPLATE_END = """
    </div>
    """
DYNAMIC_SUFFIX = """
{% endif %}
"""


class RevealDirective(RunestoneDirective):
    """
    .. reveal:: identifier
       :showtitle: Text on the 'show' button--default is "Show"
       :hidetitle: Text on the 'hide' button--default is "Hide"
       :modal: Boolean--if included, revealed display will be a modal
       :modaltitle: Title of modal dialog window--default is "Message from the author"
       :instructoronly: Only show button and contents to instructors

       Content  everything here will be hidden until revealed
       Content  It can be a lot...
    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    has_content = True
    option_spec = RunestoneDirective.option_spec.copy()
    option_spec.update(
        {
            "showtitle": directives.unchanged,
            "hidetitle": directives.unchanged,
            "modal": directives.flag,
            "modaltitle": directives.unchanged,
            "instructoronly": directives.flag,
        }
    )

    def run(self):
        """
        process the reveal directive and generate html for output.
        :param self:
        :return:
        .. reveal:: identifier
        :showtitle: Text on the 'show' button--default is "Show"
        :hidetitle: Text on the 'hide' button--default is "Hide"
        :modal: Boolean--if included, revealed display will be a modal
        :modaltitle: Title of modal dialog window--default is "Message from the author"
        :instructoronly: only reveal content to instructors

        Content
        ...
        """
        # super(RevealDirective, self).run()
        env = self.state.document.settings.env
        self.assert_has_content()  # make sure reveal has something in it

        self.options["divid"] = self.arguments[0]

        if "showtitle" not in self.options:
            self.options["showtitle"] = 'data-showtitle="Show"'
        else:
            self.options["showtitle"] = (
                """data-showtitle=""" + '"' + self.options["showtitle"] + '"'
            )
        if "hidetitle" not in self.options:
            self.options["hidetitle"] = 'data-hidetitle="Hide"'
        else:
            self.options["hidetitle"] = (
                """data-hidetitle=""" + '"' + self.options["hidetitle"] + '"'
            )

        if "instructoronly" in self.options:
            self.options[
                "instructoronly"
            ] = '''data-instructoronly style="display: none;"'''
        else:
            self.options["instructoronly"] = ""

        is_dynamic = env.config.html_context.get("dynamic_pages", False)
        self.options["is_dynamic"] = is_dynamic

        reveal_node = RevealNode()
        reveal_node["runestone_options"] = self.options
        (
            reveal_node["source"],
            reveal_node["line"],
        ) = self.state_machine.get_source_and_line(self.lineno)

        self.state.nested_parse(self.content, self.content_offset, reveal_node)

        return [reveal_node]



# ``core.py`` - Utilities supporting continuous integration tests
# ===============================================================
#
# Imports
# -------
# These are listed in the order prescribed by [PEP
# 8](http://www.python.org/dev/peps/pep-0008/#imports).
#
# ### Standard library
import asyncio
import io
import os
import subprocess
import sys
from pathlib import Path
from typing import Any, Awaitable, Callable, Literal, Sequence, cast

# ### Local imports
from dotenv import load_dotenv

# OS detection
# ------------
# This follows the [Python
# recommendations](https://docs.python.org/3/library/sys.html#sys.platform).
is_win = sys.platform == "win32"
is_linux = sys.platform.startswith("linux")
is_darwin = sys.platform == "darwin"

# Copied from the [Python
# docs](https://docs.python.org/3.5/library/platform.html#cross-platform).
is_64bits = sys.maxsize > 2**32


# Support code
# ------------
# ### xqt
# Pronounced "execute": provides a simple way to execute a system command.
def xqt(
    # Commands to run. For example, `'foo -param firstArg secondArg', 'bar |
    # grep alpha'`.
    *cmds: list[str],
    # This is passed directly to `subprocess.run`.
    check: bool = True,
    # This is passed directly to `subprocess.run`.
    shell: bool = True,
    # Optional keyword arguments to pass on to
    # [subprocess.run](https://docs.python.org/3/library/subprocess.html#subprocess.run).
    **kwargs: Any,
) -> subprocess.CompletedProcess | list[subprocess.CompletedProcess]:
    ret = []
    # Although the
    # [docs](https://docs.python.org/3/library/subprocess.html#subprocess.Popen)
    # state, "The only time you need to specify `shell=True` on Windows is when
    # the command you wish to execute is built into the shell (e.g. **dir** or
    # **copy**). You do not need `shell=True` to run a batch file or
    # console-based executable.", use `shell=True` to both allow shell commands
    # and to support simple redirection (such as `blah > nul`, instead of
    # passing `stdout=subprocess.DEVNULL` to `check_call`).
    for cmd in cmds:
        # Per
        # [SO](http://stackoverflow.com/questions/15931526/why-subprocess-stdout-to-a-file-is-written-out-of-order),
        # the `run` below will flush stdout and stderr, causing all the
        # subprocess output to appear first, followed by all the Python output
        # (such as the print statement above). So, flush the buffers to avoid
        # this.
        wd = kwargs.get("cwd", os.getcwd())
        flush_print(f"{wd}$ {cmd}")
        # Use bash instead of sh, so that `source` and other bash syntax works.
        # See
        # [subprocess.popen](https://docs.python.org/3/library/subprocess.html#subprocess.Popen).
        executable = "/bin/bash" if is_linux or is_darwin else None
        try:
            cp = subprocess.run(
                cmd, shell=shell, executable=executable, check=check, **kwargs
            )
        except subprocess.CalledProcessError as e:
            flush_print(f"{e.stderr or ''}{e.stdout or ''}")
            raise
        ret.append(cp)

    # Return a list only if there were multiple commands to execute.
    return ret[0] if len(ret) == 1 else ret


# ### env
# Transforms `os.environ["ENV_VAR"]` to the more compact `env.ENV_VAR`. Returns
# `None` if the env var doesn't exist.
class EnvType(type):
    def __getattr__(cls, name: str) -> str | None:
        return os.environ.get(name)

    def __setattr__(cls, name: str, value: Any) -> None:
        os.environ[name] = value


class env(metaclass=EnvType):
    pass


# ### pushd
# A context manager for pushd.
class pushd:
    """This is a context manager for pushd.  It will remember where you were
    and take you back there when you exit the context manager.

    :param path: The path to change to upon entering the context manager.
    :type path: str | Path
    """

    def __init__(
        self,
        # The path to change to upon entering the context manager.
        path: str | Path,
        verbose: bool = False,
    ):
        self.path = path
        self.verbose = verbose

    def __enter__(self) -> None:
        if self.verbose:
            flush_print(f"pushd {self.path}")
        self.cwd = os.getcwd()
        os.chdir(str(self.path))

    def __exit__(self, type_, value, traceback) -> Literal[False]:
        if self.verbose:
            flush_print(f"popd - returning to {self.cwd}.")
        os.chdir(str(self.cwd))
        return False


# ### chdir
def chdir(path: str | Path) -> None:
    flush_print(f"cd {path}")
    os.chdir(str(path))


# ### mkdir
def mkdir(path: str | Path, *args, **kwargs) -> None:
    flush_print(f"mkdir {path}")
    Path(path).mkdir(*args, **kwargs)


# ### load .env file
def load_project_dotenv(dotenv_path=None):
    if dotenv_path is None:
        dotenv_path = Path(".env")
    else:
        dotenv_path = Path(dotenv_path)
    if dotenv_path.exists():
        load_dotenv(dotenv_path)
        print("Loaded .env file")
    elif "RUNESTONE_PATH" in os.environ:
        env_path = Path(os.environ["RUNESTONE_PATH"]) / ".env"
        if env_path.exists():
            load_dotenv(env_path)
            print(f"Loaded .env file from {os.environ['RUNESTONE_PATH']}")


# ### flush_print
# Anything sent to `print` won't be printed until Python flushes its buffers,
# which means what CI logs report may be reflect what's actually being executed
# -- until the buffers are flushed.
def flush_print(*args: Any, **kwargs: Any) -> None:
    print(*args, **kwargs)
    # Flush both buffers, just in case there's something in either one.
    sys.stdout.flush()
    sys.stderr.flush()


# ### isfile
def isfile(f: str | Path) -> bool:
    _ = Path(f).is_file()
    flush_print(f"File {f} {'exists' if _ else 'does not exist'}.")
    return _


# ### isdir
def isdir(f: str | Path) -> bool:
    _ = Path(f).is_dir()
    flush_print(f"Directory {f} {'exists' if _ else 'does not exist'}.")
    return _


# This defines the type of an `async` function passed to `stream_command`.
SubprocessStreamerType = Callable[
    # It takes one argument: an stream produced by the subprocess (either
    # `process.stdout` or `process.stderr`). See the [docs](https://docs.python.org/3/library/asyncio-stream.html#streamreader).
    [asyncio.StreamReader],
    # It produces no output, but must be an ``async`` function.
    Awaitable[None],
]


# ### Subprocess streaming
# The following code was originally based on a post on
# [SO](https://stackoverflow.com/a/66400096/16038919).
## {
# This is for use with `stream_command`: provide it with output(s) and an
# (optional) filter, and it will stream data from from the subprocess's
# stdout/stderr to the `output_stream`(s) given below.
def subprocess_streamer(
    # Where to stream the data; typically, `sys.stdout.buffer` or
    # `sys.stderr.buffer`. Note that the asyncio subprocess `stderr` and
    # `stdout` produces bytes only, not text. So, we need to write to
    # `sys.stdout/err` as binary. Therefore, use
    # [.buffer](https://docs.python.org/3/library/io.html#io.TextIOBase.buffer).
    *output_stream: io.IOBase,
    # This optional function filters which input should be written to the
    # output(s). If not specified, all input is written to the output(s).
    filter: Callable[
        # The function takes one parameter: the line read from the process.
        [bytes],
        # It produces one output: either
        #
        # - The line to write to each `output_stream`, or
        # - a sequence of lines to write to each `output_stream`; the first
        #   element of the list is written to the first `output_stream`, etc.
        bytes | Sequence[bytes],
    ] = lambda line: line,
) -> SubprocessStreamerType:
    """
    This is for use with `stream_command`: provide it with output(s) and
    (optionally) a filter, and it will stream data from from the subprocess's
    stdout/stderr to the `output_stream`(s) given below.
    :param filter: _description_, defaults to lambdaline:line
    :type filter: _type_, optional
    :return: _description_
    :rtype: SubprocessStreamerType
    """

    async def _output_stream(
        # `async_run_command` will call this with either `process.stdout` or
        # `process.stderr`.
        input_stream: asyncio.StreamReader,
    ) -> None:
        while not input_stream.at_eof():
            output = await input_stream.readline()
            filtered_output = filter(output)
            if filtered_output:
                for index, _os in enumerate(output_stream):
                    _os.write(
                        filtered_output
                        if isinstance(filtered_output, bytes)
                        else filtered_output[index]
                    )
                    _os.flush()

    return _output_stream


# Run a subprocess and stream the output while it runs. Return the process'
# return code. This is typically called by `stream_command`.
async def async_stream_command(
    # The command to execute; same as the arguments to `subprocess`.
    *command: str,
    # Functions which stream the stdout/stderr produced as the command executes.
    # See the `SubprocessStreamerType` documentation above for more details.
    stdout_streamer: SubprocessStreamerType = subprocess_streamer(
        cast(io.IOBase, sys.stdout.buffer)
    ),
    stderr_streamer: SubprocessStreamerType = subprocess_streamer(
        cast(io.IOBase, sys.stderr.buffer)
    ),
    # These are passed directly to the underlying `subprocess` call.
    **kwargs: Any,
    # The return code produced when the subprocess exits.
) -> int:
    """Execute a command in the background and stream its output.

    :param stdout_streamer: , defaults to subprocess_streamer( cast(io.IOBase, sys.stdout.buffer) )
    :type stdout_streamer: SubprocessStreamerType, optional
    :param stderr_streamer: _description_, defaults to subprocess_streamer( cast(io.IOBase, sys.stderr.buffer) )
    :type stderr_streamer: SubprocessStreamerType, optional
    :return: _description_
    :rtype: int
    """
    # Add stdout/stderr destinations if they aren't specified.
    kwargs.setdefault("stdout", asyncio.subprocess.PIPE)
    kwargs.setdefault("stderr", asyncio.subprocess.PIPE)

    # Start the subprocess.
    process = await asyncio.create_subprocess_exec(
        *command,
        **kwargs,
    )
    # Determine what to stream.
    stream = []
    if process.stdout is not None:
        stream.append(stdout_streamer(process.stdout))
    if process.stderr is not None:
        stream.append(stderr_streamer(process.stderr))

    # Stream stdout and stderr while it runs.
    await asyncio.gather(*stream)

    # Terminate it nicely.
    #
    # `process.communicate()` will have no data to read but will close the pipes
    # that are implemented in C, whereas `process.wait()` will not.
    stdout_data, stderr_data = await process.communicate()
    assert not stdout_data and not stderr_data

    assert process.returncode is not None
    return process.returncode


# The primary function to call, which runs the above function in `asyncio`.
def stream_command(*args: str, **kwargs: Any) -> int:
    return asyncio.run(async_stream_command(*args, **kwargs))


## }
